@@ -1,5 +1,5 @@
 #!/usr/bin/env python3
-# Copyright (C) 2014 J.F.Dockes
+# Copyright (C) 2014-2020 J.F.Dockes
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 2 of the License, or
@@ -13,39 +13,53 @@
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the
 # Free Software Foundation, Inc.,
-# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
+# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
 # Recoll PDF extractor, with support for attachments
 #
 # pdftotext sometimes outputs unescaped text inside HTML text sections.
 # We try to correct.
 #
-# If pdftotext produces no text and tesseract is available, we try to
-# perform OCR. As this can be very slow and the result not always
-# good, we only do this if a file named $RECOLL_CONFDIR/ocrpdf exists
-#
-# We guess the OCR language in order of preference:
-#  - From the content of a ".ocrpdflang" file if it exists in the same
-#    directory as the PDF
-#  - From an RECOLL_TESSERACT_LANG environment variable
-#  - From the content of $RECOLL_CONFDIR/ocrpdf
-#  - Default to "eng"
-
-from __future__ import print_function
+# If pdftotext produces no text and the configuration allows it, we may try to
+# perform OCR.
 
 import os
 import sys
 import re
-import rclexecm
+import urllib.request
 import subprocess
 import tempfile
+import glob
+import traceback
 import atexit
 import signal
+
+import rclexecm
 import rclconfig
-import glob
+
+_mswindows = (sys.platform == "win32")
+
+# Can we access the poppler-glib python3 bindings ? This would allow extracting
+# text from annotations.
+# - On Ubuntu, this comes with package gir1.2-poppler-0.18
+# - On opensuse the package is named typelib-1_0-Poppler-0_18
+# (actual versions may differ of course).
+havepopplerglib = False
+try:
+    import gi
+    gi.require_version('Poppler', '0.18')
+    from gi.repository import Poppler
+    havepopplerglib = True
+except:
+    pass
 
 tmpdir = None
 
+_htmlprefix =b'''<html><head>
+<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">
+</head><body><pre>'''
+_htmlsuffix = b'''</pre></body></html>'''
+
 def finalcleanup():
     if tmpdir:
         vacuumdir(tmpdir)
@@ -82,8 +96,10 @@ class PDFExtractor:
         self.pdftk = None
         self.em = em
         self.tesseract = None
-        
-        self.pdftotext = rclexecm.which("pdftotext")
+
+        # Avoid picking up a default version on Windows, we want ours
+        if not _mswindows:
+            self.pdftotext = rclexecm.which("pdftotext")
         if not self.pdftotext:
             self.pdftotext = rclexecm.which("poppler/pdftotext")
             if not self.pdftotext:
@@ -98,6 +114,7 @@ class PDFExtractor:
         # (xmltag,rcltag) pairs
         self.extrameta = self.config.getConfParam("pdfextrameta")
         if self.extrameta:
+            self.extrametafix = self.config.getConfParam("pdfextrametafix")
             self._initextrameta()
 
         # Check if we need to escape portions of text where old
@@ -114,41 +131,29 @@ class PDFExtractor:
         except:
             pass
         
-        # See if we'll try to perform OCR. Need the commands and the
-        # either the presence of a file in the config dir (historical)
-        # or a set config variable.
-        self.ocrpossible = False
-        cf_doocr = self.config.getConfParam("pdfocr")
-        if cf_doocr or os.path.isfile(os.path.join(self.confdir, "ocrpdf")):
-            self.tesseract = rclexecm.which("tesseract")
-            if self.tesseract:
-                self.pdftoppm = rclexecm.which("pdftoppm")
-                if self.pdftoppm:
-                    self.ocrpossible = True
-                    self.maybemaketmpdir()
-        # self.em.rclog("OCRPOSSIBLE: %d" % self.ocrpossible)
-
         # Pdftk is optionally used to extract attachments. This takes
         # a hit on performance even in the absence of any attachments,
         # so it can be disabled in the configuration.
         self.attextractdone = False
         self.attachlist = []
         cf_attach = self.config.getConfParam("pdfattach")
+        cf_attach = rclexecm.configparamtrue(cf_attach)
         if cf_attach:
             self.pdftk = rclexecm.which("pdftk")
         if self.pdftk:
             self.maybemaketmpdir()
 
     def _initextrameta(self):
-        self.pdfinfo = rclexecm.which("pdfinfo")
+        if not _mswindows:
+            self.pdfinfo = rclexecm.which("pdfinfo")
         if not self.pdfinfo:
             self.pdfinfo = rclexecm.which("poppler/pdfinfo")
         if not self.pdfinfo:
             self.extrameta = None
             return
 
-        # extrameta is like "samename metanm|rclnm ..."
-        # we turn it into a list of pairs
+        # extrameta is like "metanm|rclnm ...", where |rclnm maybe absent (keep
+        # original name). Parse into a list of pairs.
         l = self.extrameta.split()
         self.extrameta = []
         for e in l:
@@ -174,10 +179,23 @@ class PDFExtractor:
             self.pdfinfo = None
             return
 
-        self.re_head = re.compile(r'<head>', re.IGNORECASE)
-        self.re_xmlpacket = re.compile(r'<\?xpacket[ 	]+begin.*\?>' +
-                                       r'(.*)' + r'<\?xpacket[ 	]+end',
+        self.re_head = re.compile(br'<head>', re.IGNORECASE)
+        self.re_xmlpacket = re.compile(br'<\?xpacket[ 	]+begin.*\?>' +
+                                       br'(.*)' + br'<\?xpacket[ 	]+end',
                                        flags = re.DOTALL)
+        global EMF
+        EMF = None
+        if self.extrametafix:
+            try:
+                import importlib.util
+                spec = importlib.util.spec_from_file_location(
+                    'pdfextrametafix', self.extrametafix)
+                EMF = importlib.util.module_from_spec(spec)
+                spec.loader.exec_module(EMF)
+            except Exception as err:
+                self.em.rclog("Import extrametafix failed: %s" % err)
+                EMF = None
+                pass
 
     # Extract all attachments if any into temporary directory
     def extractAttach(self):
@@ -189,11 +207,19 @@ class PDFExtractor:
         if not tmpdir or not self.pdftk:
             # no big deal
             return True
-
         try:
             vacuumdir(tmpdir)
-            subprocess.check_call([self.pdftk, self.filename, "unpack_files",
-                                   "output", tmpdir])
+            # Note: the java version of pdftk sometimes/often fails
+            # here with writing to stdout:
+            #    Error occurred during initialization of VM
+            #    Could not allocate metaspace: 1073741824 bytes
+            # Maybe unsufficient resources when started from Python ?
+            # In any case, the important thing is to discard the
+            # output, until we fix the error or preferably find a way
+            # to do it with poppler...
+            subprocess.check_call(
+                [self.pdftk, self.filename, "unpack_files", "output",
+                 tmpdir], stdout=sys.stderr)
             self.attachlist = sorted(os.listdir(tmpdir))
             return True
         except Exception as e:
@@ -201,118 +227,6 @@ class PDFExtractor:
             # Return true anyway, pdf attachments are no big deal
             return True
 
-    def extractone(self, ipath):
-        #self.em.rclog("extractone: [%s]" % ipath)
-        if not self.attextractdone:
-            if not self.extractAttach():
-                return (False, "", "", rclexecm.RclExecM.eofnow)
-        path = os.path.join(tmpdir, ipath)
-        if os.path.isfile(path):
-            f = open(path)
-            docdata = f.read();
-            f.close()
-        if self.currentindex == len(self.attachlist) - 1:
-            eof = rclexecm.RclExecM.eofnext
-        else:
-            eof = rclexecm.RclExecM.noteof
-        return (True, docdata, ipath, eof)
-
-
-    # Try to guess tesseract language. This should depend on the input
-    # file, but we have no general way to determine it. So use the
-    # environment and hope for the best.
-    def guesstesseractlang(self):
-        tesseractlang = ""
-
-        # First look for a language def file in the file's directory 
-        pdflangfile = os.path.join(os.path.dirname(self.filename),
-                                   b".ocrpdflang")
-        if os.path.isfile(pdflangfile):
-            tesseractlang = open(pdflangfile, "r").read().strip()
-        if tesseractlang:
-            return tesseractlang
-
-        # Then look for a global option. The normal way now that we
-        # have config reading capability in the handlers is to use the
-        # config. Then, for backwards compat, environment variable and
-        # file inside the configuration directory
-        tesseractlang = self.config.getConfParam("pdfocrlang")
-        if tesseractlang:
-            return tesseractlang
-        tesseractlang = os.environ.get("RECOLL_TESSERACT_LANG", "");
-        if tesseractlang:
-            return tesseractlang
-        pdflangfile = os.path.join(self.confdir, b"ocrpdf")
-        if os.path.isfile(pdflangfile):
-            tesseractlang = open(pdflangfile, "r").read().strip()
-        if tesseractlang:
-            return tesseractlang
-
-        # Half-assed trial to guess from LANG then default to english
-        localelang = os.environ.get("LANG", "").split("_")[0]
-        if localelang == "en":
-            tesseractlang = "eng"
-        elif localelang == "de":
-            tesseractlang = "deu"
-        elif localelang == "fr":
-            tesseractlang = "fra"
-        if tesseractlang:
-            return tesseractlang
-
-        if not tesseractlang:
-            tesseractlang = "eng"
-        return tesseractlang
-
-    # PDF has no text content and tesseract is available. Give OCR a try
-    def ocrpdf(self):
-
-        global tmpdir
-        if not tmpdir:
-            return ""
-
-        tesseractlang = self.guesstesseractlang()
-        # self.em.rclog("tesseractlang %s" % tesseractlang)
-
-        tesserrorfile = os.path.join(tmpdir, "tesserrorfile")
-        tmpfile = os.path.join(tmpdir, "ocrXXXXXX")
-
-        # Split pdf pages
-        try:
-            vacuumdir(tmpdir)
-            subprocess.check_call([self.pdftoppm, "-r", "300", self.filename,
-                                   tmpfile])
-        except Exception as e:
-            self.em.rclog("pdftoppm failed: %s" % e)
-            return ""
-
-        files = glob.glob(tmpfile + "*")
-        for f in files:
-            try:
-                out = subprocess.check_output([self.tesseract, f, f, "-l",
-                                               tesseractlang],
-                                              stderr = subprocess.STDOUT)
-            except Exception as e:
-                self.em.rclog("tesseract failed: %s" % e)
-
-            errlines = out.split(b'\n')
-            if len(errlines) > 2:
-                self.em.rclog("Tesseract error: %s" % out)
-
-        # Concatenate the result files
-        files = glob.glob(tmpfile + "*" + ".txt")
-        data = ""
-        for f in files:
-            data += open(f, "r").read()
-
-        if not data:
-            return ""
-        return '''<html><head>
-        <meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">
-        </head><body><pre>''' + \
-        self.em.htmlescape(data) + \
-        '''</pre></body></html>'''
-
-
     # pdftotext (used to?) badly escape text inside the header
     # fields. We do it here. This is not an html parser, and depends a
     # lot on the actual format output by pdftotext.
@@ -322,7 +236,7 @@ class PDFExtractor:
         inheader = False
         inbody = False
         didcs = False
-        output = b''
+        output = []
         isempty = True
         for line in input.split(b'\n'):
             if re.search(b'</head>', line):
@@ -331,15 +245,15 @@ class PDFExtractor:
                 inbody = False
             if inheader:
                 if not didcs:
-                    output += b'<meta http-equiv="Content-Type"' + \
-                              b'content="text/html; charset=UTF-8">\n'
+                    output.append(b'<meta http-equiv="Content-Type"' + \
+                              b'content="text/html; charset=UTF-8">\n')
                     didcs = True
                 if self.needescape:
                     m = re.search(b'''(.*<title>)(.*)(<\/title>.*)''', line)
                     if not m:
                         m = re.search(b'''(.*content=")(.*)(".*/>.*)''', line)
                     if m:
-                        line = m.group(1) + self.em.htmlescape(m.group(2)) + \
+                        line = m.group(1) + rclexecm.htmlescape(m.group(2)) + \
                                m.group(3)
 
                 # Recoll treats "Subject" as a "title" element
@@ -349,34 +263,34 @@ class PDFExtractor:
 
             elif inbody:
                 s = line[0:1]
-                if s != "\x0c" and s != "<":
+                if s != b"\x0c" and s != b"<":
                     isempty = False
                 # We used to remove end-of-line hyphenation (and join
                 # lines), but but it's not clear that we should do
                 # this as pdftotext without the -layout option does it ?
-                line = self.em.htmlescape(line)
+                line = rclexecm.htmlescape(line)
 
             if re.search(b'<head>', line):
                 inheader = True
             if re.search(b'<pre>', line):
                 inbody = True
 
-            output += line + b'\n'
+            output.append(line)
 
-        return output, isempty
+        return b'\n'.join(output), isempty
 
     def _metatag(self, nm, val):
-        return "<meta name=\"" + nm + "\" content=\"" + \
-               self.em.htmlescape(val) + "\">"
+        return b"<meta name=\"" + rclexecm.makebytes(nm) + b"\" content=\"" + \
+               rclexecm.htmlescape(rclexecm.makebytes(val)) + b"\">"
 
     # metaheaders is a list of (nm, value) pairs
     def _injectmeta(self, html, metaheaders):
-        metatxt = ''
+        metatxt = b''
         for nm, val in metaheaders:
-            metatxt += self._metatag(nm, val) + '\n'
+            metatxt += self._metatag(nm, val) + b'\n'
         if not metatxt:
             return html
-        res = self.re_head.sub('<head>\n' + metatxt, html)
+        res = self.re_head.sub(b'<head>\n' + metatxt, html)
         #self.em.rclog("Substituted html: [%s]"%res)
         if res:
             return res
@@ -395,15 +309,14 @@ class PDFExtractor:
         
     def _setextrameta(self, html):
         if not self.pdfinfo:
-            return
+            return html
 
+        emf = EMF.MetaFixer() if EMF else None
+
+        # Execute pdfinfo and extract the XML packet
         all = subprocess.check_output([self.pdfinfo, "-meta", self.filename])
-
-        # Extract the XML packet
         res = self.re_xmlpacket.search(all)
-        xml = ''
-        if res:
-            xml = res.group(1)
+        xml = res.group(1) if res else ''
         # self.em.rclog("extrameta: XML: [%s]" % xml)
         if not xml:
             return html
@@ -417,25 +330,140 @@ class PDFExtractor:
         namespaces = {'rdf' : "http://www.w3.org/1999/02/22-rdf-syntax-ns#"}
         rdf = root.find("rdf:RDF", namespaces)
         #self.em.rclog("RDF NSMAP: %s"% rdf.nsmap)
+        if rdf is None:
+            return html
         rdfdesclist = rdf.findall("rdf:Description", rdf.nsmap)
-        #self.em.rclog("RDFDESC NSMAP: %s"% rdfdesc.nsmap)
         for metanm,rclnm in self.extrameta:
             for rdfdesc in rdfdesclist:
                 try:
-                    elt = rdfdesc.find(metanm, rdfdesc.nsmap)
+                    elts = rdfdesc.findall(metanm, rdfdesc.nsmap)
                 except:
                     # We get an exception when this rdf:Description does not
                     # define the required namespace.
                     continue
-                if elt is not None:
-                    text = self._xmltreetext(elt)
+                
+                if elts:
+                    for elt in elts:
+                        text = None
+                        try:
+                            # First try to get text from a custom element handler
+                            text = emf.metafixelt(metanm, elt)
+                        except:
+                            pass
+                        
+                        if text is None:
+                            # still nothing here, read the element text
+                            text = self._xmltreetext(elt)
+                            try:
+                                # try to run metafix
+                                text = emf.metafix(metanm, text)
+                            except:
+                                pass
+
+                        if text:
+                            # Can't use setfield as it only works for
+                            # text/plain output at the moment.
+                            #self.em.rclog("Appending: (%s,%s)"%(rclnm,text))
+                            metaheaders.append((rclnm, text))
+                else:
+                    # Some docs define the values as attributes. don't
+                    # know if this is valid but anyway...
+                    try:
+                        prefix,nm = metanm.split(":")
+                        fullnm = "{%s}%s" % (rdfdesc.nsmap[prefix], nm)
+                    except:
+                        fullnm = metanm
+                    text = rdfdesc.get(fullnm)
                     if text:
-                        # Should we set empty values ?
-                        # Can't use setfield as it only works for
-                        # text/plain output at the moment.
+                        try:
+                            # try to run metafix
+                            text = emf.metafix(metanm, text)
+                        except:
+                            pass
                         metaheaders.append((rclnm, text))
         if metaheaders:
+            if emf:
+                try:
+                    emf.wrapup(metaheaders)
+                except:
+                    pass
             return self._injectmeta(html, metaheaders)
+        else:
+            return html
+
+    def maybemaketmpdir(self):
+        global tmpdir
+        if tmpdir:
+            if not vacuumdir(tmpdir):
+                self.em.rclog("openfile: vacuumdir %s failed" % tmpdir)
+                return False
+        else:
+            tmpdir = tempfile.mkdtemp(prefix='rclmpdf')
+            if self.pdftk and re.match("/snap/", self.pdftk):
+                # We know this is Unix (Ubuntu actually). Check that tmpdir
+                # belongs to the user as snap commands can't use /tmp to share
+                # files. Don't generate an error as this only affects
+                # attachment extraction
+                ok = False
+                if "TMPDIR" in os.environ:
+                    st = os.stat(os.environ["TMPDIR"])
+                    if st.st_uid == os.getuid():
+                        ok = True
+                if not ok:
+                    self.em.rclog(
+                        "pdftk is a snap command and needs TMPDIR to be "
+                        "a directory you own")
+
+    def _process_annotations(self, html):
+        doc = Poppler.Document.new_from_file(
+            'file://%s' %
+            urllib.request.pathname2url(os.path.abspath(self.filename)), None)
+        n_pages = doc.get_n_pages()
+        all_annots = 0
+
+        # output format
+        f = 'P.: {0}, {1:10}, {2:10}: {3}'
+
+        # Array of annotations indexed by page number. The page number
+        # here is the physical one (evince -i), not a page label (evince
+        # -p). This may be different for some documents.
+        abypage = {}
+        for i in range(n_pages):
+            page = doc.get_page(i)
+            pnum = i+1
+            annot_mappings = page.get_annot_mapping ()
+            num_annots = len(annot_mappings)
+            for annot_mapping in annot_mappings:
+                atype = annot_mapping.annot.get_annot_type().value_name
+                if atype  != 'POPPLER_ANNOT_LINK':
+                    # Catch because we sometimes get None values
+                    try:
+                        atext = f.format(
+                            pnum,
+                            annot_mapping.annot.get_modified(),
+                            annot_mapping.annot.get_annot_type().value_nick,
+                            annot_mapping.annot.get_contents()) + "\n"
+                        if pnum in abypage:
+                            abypage[pnum] += atext
+                        else:
+                            abypage[pnum] = atext
+                    except:
+                        pass
+        #self.em.rclog("Annotations: %s" % abypage)
+        pagevec = html.split(b"\f")
+        html = b""
+        annotsfield = b""
+        pagenum = 1
+        for page in pagevec:
+            html += page
+            if pagenum in abypage:
+                html += abypage[pagenum].encode('utf-8')
+                annotsfield += abypage[pagenum].encode('utf-8') + b" - " 
+            html += b"\f"
+            pagenum += 1
+        if annotsfield:
+            self.em.setfield("pdfannot", annotsfield)
+        return html
     
     def _selfdoc(self):
         '''Extract the text from the pdf doc (as opposed to attachment)'''
@@ -453,25 +481,51 @@ class PDFExtractor:
         html, isempty = self._fixhtml(html)
         #self.em.rclog("ISEMPTY: %d : data: \n%s" % (isempty, html))
 
-        if isempty and self.ocrpossible:
-            html = self.ocrpdf()
+        if isempty:
+            self.config.setKeyDir(os.path.dirname(self.filename))
+            s = self.config.getConfParam("pdfocr")
+            if rclexecm.configparamtrue(s):
+                try:
+                    cmd = [sys.executable, os.path.join(_execdir, "rclocr.py"),
+                           self.filename]
+                    data = subprocess.check_output(cmd)
+                    html = _htmlprefix + rclexecm.htmlescape(data) + _htmlsuffix
+                except Exception as e:
+                    self.em.rclog("%s failed: %s" % (cmd, e))
+                    pass
 
         if self.extrameta:
             try:
                 html = self._setextrameta(html)
             except Exception as err:
-                self.em.rclog("Metadata extraction failed: %s" % err)
+                self.em.rclog("Metadata extraction failed: %s %s" %
+                              (err, traceback.format_exc()))
 
+        if havepopplerglib:
+            try:
+                html = self._process_annotations(html)
+            except Exception as err:
+                self.em.rclog("Annotation extraction failed: %s %s" %
+                              (err, traceback.format_exc()))
         return (True, html, "", eof)
 
-    def maybemaketmpdir(self):
-        global tmpdir
-        if tmpdir:
-            if not vacuumdir(tmpdir):
-                self.em.rclog("openfile: vacuumdir %s failed" % tmpdir)
-                return False
+
+    def extractone(self, ipath):
+        #self.em.rclog("extractone: [%s]" % ipath)
+        if not self.attextractdone:
+            if not self.extractAttach():
+                return (False, "", "", rclexecm.RclExecM.eofnow)
+        path = os.path.join(tmpdir, ipath)
+        if os.path.isfile(path):
+            f = open(path, "rb")
+            docdata = f.read();
+            f.close()
+        if self.currentindex == len(self.attachlist) - 1:
+            eof = rclexecm.RclExecM.eofnext
         else:
-            tmpdir = tempfile.mkdtemp(prefix='rclmpdf')
+            eof = rclexecm.RclExecM.noteof
+        return (True, docdata, ipath, eof)
+
         
     ###### File type handler api, used by rclexecm ---------->
     def openfile(self, params):
@@ -479,7 +533,7 @@ class PDFExtractor:
             print("RECFILTERROR HELPERNOTFOUND pdftotext")
             sys.exit(1);
 
-        self.filename = rclexecm.subprocfile(params["filename:"])
+        self.filename = rclexecm.subprocfile(params["filename"])
 
         #self.em.rclog("openfile: [%s]" % self.filename)
         self.currentindex = -1
@@ -498,7 +552,7 @@ class PDFExtractor:
         return True
 
     def getipath(self, params):
-        ipath = params["ipath:"]
+        ipath = params["ipath"]
         ok, data, ipath, eof = self.extractone(ipath)
         return (ok, data, ipath, eof)
         
@@ -523,11 +577,14 @@ class PDFExtractor:
 
                 #self.em.rclog("getnext: returning ok for [%s]" % ipath)
                 return (ok, data, ipath, eof)
-            except:
+            except Exception as ex:
+                self.em.rclog("getnext: extractone failed for index %d: %s" %
+                                  (self.currentindex, ex))
                 return (False, "", "", rclexecm.RclExecM.eofnow)
 
 
 # Main program: create protocol handler and extractor and run them
+_execdir = os.path.dirname(sys.argv[0])
 proto = rclexecm.RclExecM()
 extract = PDFExtractor(proto)
 rclexecm.main(proto, extract)
